downloadable_dataset_urls = {
    "ag-raw-train": {
        "filename": "train.csv",
        "url": ("https://raw.githubusercontent.com/mhjabreel/CharCnn_Keras/master/"
                "data/ag_news_csv/train.csv"),
        "md5": "b1a00f826fdfbd249f79597b59e1dc12",
        "untar": False,
        "unzip": False,
    },
    "ag-raw-test": {
        "filename": "test.csv",
        "url": ("https://raw.githubusercontent.com/mhjabreel/CharCnn_Keras/master/data/"
                "ag_news_csv/test.csv"),
        "md5": "d52ea96a97a2d943681189a97654912d",
        "untar": False,
        "unzip": False,
    },
    "imdb-raw": {
        "filename": "aclImdb_v1.tar.gz",
        "url": "https://ai.stanford.edu/~amaas/data/sentiment/aclImdb_v1.tar.gz",
        "md5": "7c2ac02c03563afcf9b574c7e56c153a",
        "untar": True,
        "unzip": False,
    },
    "yelp-raw": {
        "filename": "yelp_review_polarity_csv.tgz",
        "url": "https://s3.amazonaws.com/fast-ai-nlp/yelp_review_polarity_csv.tgz",
        "md5": "0f09b3af1a79c136ef9ca5f29df9ed9a",
        "untar": True,
        "unzip": False,
    },
    "mr-raw": {
        "filename": "rt-polaritydata.tar.gz",
        "url": "http://www.cs.cornell.edu/people/pabo/movie-review-data/rt-polaritydata.tar.gz",
        "md5": "50c1c2c047b4225e148e97aa7708c34e",
        "untar": True,
        "unzip": False,
    },
    "snli-raw": {
        "filename": "snli_1.0.zip",
        "url": "https://nlp.stanford.edu/projects/snli/snli_1.0.zip",
        "md5": "981c3df556bbaea3f17f752456d0088c",
        "untar": False,
        "unzip": True,
    },
    "mnli-raw": {
        "filename": "multinli_1.0.zip",
        "url": "https://cims.nyu.edu/~sbowman/multinli/multinli_1.0.zip",
        "md5": "0f70aaf66293b3c088a864891db51353",
        "untar": False,
        "unzip": True,
    },
    "processed-datasets": {
        "filename": "datasets.tgz",
        "url": "https://fibber-data.s3.amazonaws.com/datasets_v0.2.tgz",
        "md5": "81b1eda8421764a0b648cb83425f53a3",
        "untar": True,
        "unzip": False,
    },
    "mr-demo": {
        "filename": "mr-demo.tgz",
        "url": "https://fibber-data.s3.amazonaws.com/mr-demo.tgz",
        "md5": "8ee6b6eda12a7c1282cb903c713085c5",
        "untar": True,
        "unzip": False,
    }
}
