downloadable_resource_urls = {
    "default-glove-embeddings": {
        "filename": "glove.tgz",
        "url": "https://fibber-data.s3.amazonaws.com/glove.tgz",
        "md5": "7deac3ce2cfd8be24f97bc4ba007a19c",
        "untar": True,
        "unzip": False,
    },
    "default-stopwords": {
        "filename": "stopwords.txt",
        "url": "https://fibber-data.s3.amazonaws.com/stopwords_v1.txt",
        "md5": "6d3c778d216390063f424112087e5c10",
        "untar": False,
        "unzip": False,
    },
    "nltk-punkt": {
        "filename": "punkt.zip",
        "url": "https://fibber-data.s3.amazonaws.com/punkt.zip",
        "md5": "398bbed6dd3ebb0752fe0735d1c418fe",
        "untar": False,
        "unzip": True,
    },
    "nltk_stopwords": {
        "filename": "stopwords.zip",
        "url": "https://fibber-data.s3.amazonaws.com/nltk_stopwords.zip",
        "md5": "884694b9055d1caee8a0ca3aa3b2c7f7",
        "untar": False,
        "unzip": True,
    },
    "universal-sentence-encoder": {
        "filename": "universal-sentence-encoder-large_5.tar.gz",
        "url": "https://fibber-data.s3.amazonaws.com/universal-sentence-encoder-large_5.tar.gz",
        "md5": "1ffeec98147fd21cc2d5627ec09f0d57",
        "untar": True,
        "unzip": False,
    },
    "bert-base-cased": {
        "filename": "bert-base-cased.tgz",
        "url": "https://fibber-data.s3.amazonaws.com/bert-base-cased.tgz",
        "md5": "24e9ed53ff28518793de40c344ba0b6a",
        "untar": True,
        "unzip": False,
    },
    "bert-base-uncased": {
        "filename": "bert-base-uncased.tgz",
        "url": "https://fibber-data.s3.amazonaws.com/bert-base-uncased.tgz",
        "md5": "0c41b10e7dd3a42cfd314e2bab315b25",
        "untar": True,
        "unzip": False,
    },
    "gpt2-medium": {
        "filename": "gpt2-medium.tgz",
        "url": "https://fibber-data.s3.amazonaws.com/gpt2-medium.tgz",
        "md5": "2de5082b58bc7a2d2fe4f744db2fdafc",
        "untar": True,
        "unzip": False,
    },
    "stanford-corenlp": {
        "filename": "stanford-corenlp-4.1.0.zip",
        "url": "https://fibber-data.s3.amazonaws.com/stanford-corenlp-4.1.0.zip",
        "md5": "2ee2e614eec7f913800957297139f63d",
        "untar": False,
        "unzip": True,
    },
    "bert-base-uncased-clf-demo": {
        "filename": "bert-base-uncased-clf-demo.tgz",
        "url": "https://fibber-data.s3.amazonaws.com/bert-base-uncased-clf-demo.tgz",
        "md5": "4af6ca998625bbd64d9a6d59a3af035f",
        "untar": True,
        "unzip": False,
    },
    "bert-base-uncased-lm-demo": {
        "filename": "bert-base-uncased-lm-demo.tgz",
        "url": "https://fibber-data.s3.amazonaws.com/bert-base-uncased-lm-demo.tgz",
        "md5": "ff8f302c81b0aace15890842b7bb0e17",
        "untar": True,
        "unzip": False
    },
    "wpe-demo": {
        "filename": "wpe-demo.tgz",
        "url": "https://fibber-data.s3.amazonaws.com/wpe-demo.tgz",
        "md5": "673fb249c2d63e334bc224c4a7f9bc89",
        "untar": True,
        "unzip": False
    }
}
